@@ -1,4 +1,8 @@
 # Copyright Toyota Research Institute 2019
+"""
+Contains agents related to crystal-structure discovery
+and hypothesizing
+"""
 
 import time
 import abc
@@ -8,6 +12,7 @@ from copy import deepcopy
 from multiprocessing import cpu_count
 from collections import OrderedDict
 
+import tensorflow as tf
 import gpflow
 import numpy as np
 from qmpy.analysis.thermodynamics.phase import Phase, PhaseData
@@ -35,8 +40,15 @@ class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
     initialization parameters and methods common to every agent
     which is responsible for making decisions about stability.
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count()):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -68,9 +80,15 @@ class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
             None
         """
         self.pd = PhaseData()
-        phases = [Phase(row['Composition'], energy=row['delta_e'],
-                        per_atom=True, description=row_index)
-                  for row_index, row in self.seed_data.iterrows()]
+        phases = [
+            Phase(
+                row["Composition"],
+                energy=row["delta_e"],
+                per_atom=True,
+                description=row_index,
+            )
+            for row_index, row in self.seed_data.iterrows()
+        ]
         phases.extend([Phase(el, 0.0, per_atom=True) for el in ELEMENTS])
         self.pd.add_phases(phases)
         return self.pd
@@ -95,27 +113,31 @@ class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
         # Note: In the drop command, we're ignoring errors for
         #   brevity.  We should watch this, because we may not
         #   drop everything we intend to.
-        drop_columns = ['Composition', 'N_species', 'delta_e',
-                        'pred_delta_e', 'pred_stability',
-                        'stability', 'is_stable', 'structure']
+        drop_columns = [
+            "Composition",
+            "N_species",
+            "delta_e",
+            "pred_delta_e",
+            "pred_stability",
+            "stability",
+            "is_stable",
+            "structure",
+        ]
         if candidate_data is not None:
             self.candidate_data = candidate_data
-            X_cand = candidate_data.drop(
-                drop_columns, axis=1, errors='ignore')
+            X_cand = candidate_data.drop(drop_columns, axis=1, errors="ignore")
         else:
             X_cand = None
         if seed_data is not None:
             self.seed_data = seed_data
-            X_seed = self.seed_data.drop(
-                drop_columns, axis=1, errors='ignore')
-            y_seed = self.seed_data['delta_e']
+            X_seed = self.seed_data.drop(drop_columns, axis=1, errors="ignore")
+            y_seed = self.seed_data["delta_e"]
         else:
             X_seed, y_seed = None, None
 
         return X_cand, X_seed, y_seed
 
-    def update_candidate_stabilities(self, formation_energies,
-                                     sort=True, floor=-6.0):
+    def update_candidate_stabilities(self, formation_energies, sort=True, floor=-6.0):
         """
         Updates the candidate dataframe with the stabilities
         of the candidate compositions according to the requisite
@@ -137,12 +159,16 @@ class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
             formation_energies[formation_energies < floor] = floor
 
         # Update formation energy predictions
-        self.candidate_data['pred_delta_e'] = formation_energies
+        self.candidate_data["pred_delta_e"] = formation_energies
 
         # Construct candidate phases
         candidate_phases = [
-            Phase(data['Composition'], energy=data['pred_delta_e'],
-                  per_atom=True, description=m_id)
+            Phase(
+                data["Composition"],
+                energy=data["pred_delta_e"],
+                per_atom=True,
+                description=m_id,
+            )
             for m_id, data in self.candidate_data.iterrows()
         ]
 
@@ -153,20 +179,32 @@ class StabilityAgent(HypothesisAgent, metaclass=abc.ABCMeta):
 
         # Compute and return stabilities
         space_ml.compute_stabilities(candidate_phases, self.parallel)
-        self.candidate_data['pred_stability'] = \
-            [phase.stability for phase in candidate_phases]
+        self.candidate_data["pred_stability"] = [
+            phase.stability for phase in candidate_phases
+        ]
 
         if sort:
-            self.candidate_data = self.candidate_data.sort_values(
-                'pred_stability')
+            self.candidate_data = self.candidate_data.sort_values("pred_stability")
 
         return self.candidate_data
 
 
 class QBCStabilityAgent(StabilityAgent):
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(), alpha=0.5,
-                 training_fraction=0.5, model=None, n_members=10):
+    """
+    Agent which uses QBC to determine optimal hypotheses
+    """
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        alpha=0.5,
+        training_fraction=0.5,
+        model=None,
+        n_members=10,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -185,21 +223,35 @@ class QBCStabilityAgent(StabilityAgent):
         """
 
         super(QBCStabilityAgent, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
 
         self.alpha = alpha
         self.model = model
         self.n_members = n_members
         self.qbc = QBC(
-            n_members=n_members, training_fraction=training_fraction,
-            model=model,
+            n_members=n_members, training_fraction=training_fraction, model=model,
         )
 
-    def get_hypotheses(self, candidate_data, seed_data=None,
-                       retrain_committee=True):
+    def get_hypotheses(self, candidate_data, seed_data=None, retrain_committee=True):
+        """
+        Get hypotheses method for QBCStabilityAgent
+
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit GPUCB
+            retrain_committee (bool): whether to retrain committee
+                each time
+
+        Returns:
+            (pandas.DataFrame): top candidates from the GPUCB algorithm
+
+        """
         X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
 
         # Retrain committee if untrained or if specified
@@ -212,11 +264,10 @@ class QBCStabilityAgent(StabilityAgent):
         expected = preds - stds * self.alpha
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Find the most stable ones up to n_query within hull_distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
 
         return within_hull.head(self.n_query)
@@ -228,9 +279,17 @@ class AgentStabilityML5(StabilityAgent):
     exploitation in each iteration.  It will exploit a fraction of
     N_query options (frac), and explore the rest of its budget.
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(),
-                 model=None, exploit_fraction=0.5):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        model=None,
+        exploit_fraction=0.5,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -245,38 +304,57 @@ class AgentStabilityML5(StabilityAgent):
                 exploitation hypotheses
         """
         super(AgentStabilityML5, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
 
         self.model = model or LinearRegression()
         self.exploit_fraction = exploit_fraction
 
     def get_hypotheses(self, candidate_data, seed_data=None):
+        """
+        Get hypotheses method for AgentStabilityML5
+
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit GPUCB
+
+        Returns:
+            (pandas.DataFrame): top candidates from the GPUCB algorithm
+
+        """
+
         X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
-        steps = [('scaler', StandardScaler()), ('ML', self.model)]
+        steps = [("scaler", StandardScaler()), ("ML", self.model)]
         pipeline = Pipeline(steps)
 
-        cv_score = cross_val_score(pipeline, X_seed, self.seed_data['delta_e'],
-                                   cv=KFold(5, shuffle=True), scoring='neg_mean_absolute_error')
-        self.cv_score = np.mean(cv_score)*-1
-        pipeline.fit(X_seed, self.seed_data['delta_e'])
+        cv_score = cross_val_score(
+            pipeline,
+            X_seed,
+            self.seed_data["delta_e"],
+            cv=KFold(5, shuffle=True),
+            scoring="neg_mean_absolute_error",
+        )
+        self.cv_score = np.mean(cv_score) * -1
+        pipeline.fit(X_seed, self.seed_data["delta_e"])
 
         expected = pipeline.predict(X_cand)
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Filter by stability according to hull distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
 
         # Exploitation part:
         n_exploitation = int(self.n_query * self.exploit_fraction)
         to_compute = within_hull.head(n_exploitation).index.tolist()
-        remaining = within_hull.tail(len(within_hull)- n_exploitation)
+        remaining = within_hull.tail(len(within_hull) - n_exploitation)
         remaining = remaining.append(self.candidate_data[~stability_filter])
 
         # Exploration part (pick randomly from remainder):
@@ -292,9 +370,16 @@ class GaussianProcessStabilityAgent(StabilityAgent):
     """
     Simple Gaussian Process Regressor based Stability Agent
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(),
-                 alpha=0.5):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        alpha=0.5,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -308,24 +393,50 @@ class GaussianProcessStabilityAgent(StabilityAgent):
                 best-case predictions of the stability
         """
         super(GaussianProcessStabilityAgent, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
         self.multiprocessing = parallel
         self.alpha = alpha
-        self.GP = GaussianProcessRegressor(kernel=ConstantKernel(1) * RBF(1), alpha=0.002)
+        self.GP = GaussianProcessRegressor(
+            kernel=ConstantKernel(1) * RBF(1), alpha=0.002
+        )
 
     def get_hypotheses(self, candidate_data, seed_data=None):
-        X_cand, X_seed, y_seed = self.update_data(candidate_data=candidate_data, seed_data=seed_data)
+        """
+        Get hypotheses method for GaussianProcessStabilityAgent
 
-        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit GPUCB
+
+        Returns:
+            (pandas.DataFrame): top candidates from the GPUCB algorithm
+
+        """
+
+        X_cand, X_seed, y_seed = self.update_data(
+            candidate_data=candidate_data, seed_data=seed_data
+        )
+
+        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
         cv_pipeline = Pipeline(steps)
         self.cv_score = np.mean(
-            -1.0 * cross_val_score(cv_pipeline, X_seed, y_seed, cv=KFold(3, shuffle=True),
-                                   scoring='neg_mean_absolute_error'))
+            -1.0
+            * cross_val_score(
+                cv_pipeline,
+                X_seed,
+                y_seed,
+                cv=KFold(3, shuffle=True),
+                scoring="neg_mean_absolute_error",
+            )
+        )
 
-        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
+        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
         overall_pipeline = Pipeline(steps)
         overall_pipeline.fit(X_seed, y_seed)
 
@@ -334,11 +445,10 @@ class GaussianProcessStabilityAgent(StabilityAgent):
         expected = preds - stds * self.alpha
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Find the most stable ones up to n_query within hull_distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
 
         return within_hull.head(self.n_query)
@@ -367,9 +477,17 @@ class SVGProcessStabilityAgent(StabilityAgent):
             optimization." arXiv preprint arXiv:1412.6980 (2014).
 
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(),
-                 alpha=0.5, M=600):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        alpha=0.5,
+        M=600,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -385,9 +503,11 @@ class SVGProcessStabilityAgent(StabilityAgent):
                 SVGProcess
         """
         super(SVGProcessStabilityAgent, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
         self.alpha = alpha
         self.M = M
@@ -401,11 +521,30 @@ class SVGProcessStabilityAgent(StabilityAgent):
         self.pred_std = None
 
     def get_hypotheses(self, candidate_data, seed_data=None):
+        """
+        Get hypotheses method for SVGProcessStabilityAgent.
+
+        Code used from gpflow examples for big data, see:
+
+        https://github.com/GPflow/GPflow/blob/develop/doc/source/notebooks/advanced/gps_for_big_data.pct.py
+
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit model
+
+        Returns:
+            (pandas.DataFrame): top candidates from the algorithm
+
+        """
+
         X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
 
         # Test model performance first.  Note we avoid doing CV to
         # reduce compute time. We simply do a 1-way split 80:20 (train:test)
-        X_train, X_test, y_train, y_test = train_test_split(X_seed, y_seed, test_size=0.2, random_state=42)
+        X_train, X_test, y_train, y_test = train_test_split(
+            X_seed, y_seed, test_size=0.2, random_state=42
+        )
         scaler = StandardScaler()
         X_train_scaled = scaler.fit_transform(X_train)
 
@@ -419,14 +558,50 @@ class SVGProcessStabilityAgent(StabilityAgent):
         print(Z, _y.shape)
         print(sig, mu)
         model = gpflow.models.SVGP(
-            X_train_scaled, ((_y - mu)/sig).reshape(-1, 1), self.kernel,
-            gpflow.likelihoods.Gaussian(), Z, mean_function=self.mean_f,
-            minibatch_size=100
+            self.kernel,
+            gpflow.likelihoods.Gaussian(),
+            Z,
+            mean_function=self.mean_f,
         )
+        minibatch_size = 100
+
+        # We turn off training for inducing point locations
+        gpflow.utilities.set_trainable(model.inducing_variable, False)
+
+        train_dataset = tf.data.Dataset.from_tensor_slices((X_train_scaled, y_train)) \
+            .repeat() \
+            .shuffle(len(X_train_scaled))
+
+        @tf.function(autograph=False)
+        def optimization_step(optimizer, model: gpflow.models.SVGP, batch):
+            with tf.GradientTape(watch_accessed_variables=False) as tape:
+                tape.watch(model.trainable_variables)
+                objective = - model.elbo(batch)
+                grads = tape.gradient(objective, model.trainable_variables)
+            optimizer.apply_gradients(zip(grads, model.trainable_variables))
+            return objective
+
+        def run_adam(model, iterations):
+            """
+            Utility function running the Adam optimizer
+
+            :param model: GPflow model
+            :param interations: number of iterations
+            """
+            # Create an Adam Optimizer action
+            logf = []
+            train_it = iter(train_dataset.batch(minibatch_size))
+            adam = tf.optimizers.Adam()
+            for step in range(iterations):
+                elbo = - optimization_step(adam, model, next(train_it))
+                if step % 10 == 0:
+                    logf.append(elbo.numpy())
+            return logf
+
         print("training")
         t0 = time.time()
-        logger = self.run_adam(model, gpflow.test_util.notebook_niter(20000))
-        print("elapsed time: ", time.time()-t0)
+        run_adam(model, gpflow.ci_utils.ci_niter(20000))
+        print("elapsed time: ", time.time() - t0)
 
         pred_y, pred_v = model.predict_y(scaler.transform(X_test))
         pred_y = pred_y * sig + mu
@@ -438,17 +613,19 @@ class SVGProcessStabilityAgent(StabilityAgent):
         scaler = StandardScaler()
         X_scaled = scaler.fit_transform(X_seed)
         cls = MiniBatchKMeans(n_clusters=self.M, batch_size=200)
-        cls.fit(X_train_scaled)
+        cls.fit(X_scaled)
         Z = cls.cluster_centers_
         _y = np.array(y_seed.to_list())
         mu = np.mean(_y)
         sig = np.std(_y)
         model = gpflow.models.SVGP(
-            X_scaled, ((_y - mu)/sig).reshape(-1, 1), self.kernel,
-            gpflow.likelihoods.Gaussian(), Z, mean_function=self.mean_f,
-            minibatch_size=100
+            self.kernel,
+            gpflow.likelihoods.Gaussian(),
+            Z,
+            mean_function=self.mean_f,
         )
-        logger = self.run_adam(model, gpflow.test_util.notebook_niter(20000))
+        maxiter = gpflow.ci_utils.ci_niter(20000)
+        self.run_adam(model, maxiter)
         print(self.model)
         self.model = model
 
@@ -456,57 +633,20 @@ class SVGProcessStabilityAgent(StabilityAgent):
         pred_y, pred_v = model.predict_y(scaler.transform(X_cand))
         pred_y = pred_y * sig + mu
         self.pred_y = pred_y.reshape(-1,)
-        self.pred_std = (pred_v**0.5).reshape(-1,)
+        self.pred_std = (pred_v ** 0.5).reshape(-1,)
 
         expected = self.pred_y - self.pred_std * self.alpha
         print("expected improv", expected)
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Find the most stable ones up to n_query within hull_distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
 
         return within_hull.head(self.n_query)
 
-    def run_adam(self, model, iterations):
-        """
-        Adam optimizer as implemented in:
-        https://github.com/GPflow/GPflow/blob/develop/doc/source/notebooks/advanced/gps_for_big_data.ipynb
-        """
-        # Create an Adam Optimiser action
-        adam = gpflow.train.AdamOptimizer().make_optimize_action(model)
-
-        # Create a Logger action
-        self.logger = self.Logger(model)
-        actions = [adam, self.logger]
-
-        # Create optimisation loop that interleaves Adam with Logger
-        loop = gpflow.actions.Loop(actions, stop=iterations)()
-
-        # Bind current TF session to model
-        model.anchor(model.enquire_session())
-        return self.logger
-
-    class Logger(gpflow.actions.Action):
-        """
-        Logger class as implemented in
-        https://github.com/GPflow/GPflow/blob/develop/doc/source/notebooks/advanced/gps_for_big_data.ipynb
-        """
-        def __init__(self, model):
-            self.model = model
-            self.logf = []
-
-        def run(self, ctx):
-            if (ctx.iteration % 10) == 0:
-                # Extract likelihood tensor from Tensorflow session
-                likelihood = - ctx.session.run(self.model.likelihood_tensor)
-
-                # Append likelihood value to list
-                self.logf.append(likelihood)
-
 
 class BaggedGaussianProcessStabilityAgent(StabilityAgent):
     """
@@ -522,10 +662,19 @@ class BaggedGaussianProcessStabilityAgent(StabilityAgent):
     the minimum std among all GP estimators trained, for a given point.
 
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(),
-                 alpha=0.5, n_estimators=8, max_samples=5000,
-                 bootstrap=False):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        alpha=0.5,
+        n_estimators=8,
+        max_samples=5000,
+        bootstrap=False,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -539,9 +688,11 @@ class BaggedGaussianProcessStabilityAgent(StabilityAgent):
                 case formation energies
         """
         super(BaggedGaussianProcessStabilityAgent, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
 
         self.alpha = alpha
@@ -549,22 +700,46 @@ class BaggedGaussianProcessStabilityAgent(StabilityAgent):
         self.max_samples = max_samples if max_samples else 5000
         self.bootstrap = bootstrap if bootstrap else False
 
-        self.GP = GaussianProcessRegressor(kernel=ConstantKernel(1) * RBF(1), alpha=0.002)
+        self.GP = GaussianProcessRegressor(
+            kernel=ConstantKernel(1) * RBF(1), alpha=0.002
+        )
 
     def get_hypotheses(self, candidate_data, seed_data=None):
+        """
+        Get hypotheses method for BaggedGaussianProcessStabilityAgent
+
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit model
+
+        Returns:
+            (pandas.DataFrame): top candidates from the algorithm
+
+        """
+
         X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
 
-        steps = [('scaler', StandardScaler()), ('GP', self.GP)]
+        steps = [("scaler", StandardScaler()), ("GP", self.GP)]
         pipeline = Pipeline(steps)
         n_jobs = self.parallel if self.parallel else None
         bag_reg = BaggingRegressor(
-            base_estimator=pipeline, n_estimators=self.n_estimators,
-            max_samples=self.max_samples, bootstrap=self.bootstrap,
-            verbose=True, n_jobs=n_jobs)
+            base_estimator=pipeline,
+            n_estimators=self.n_estimators,
+            max_samples=self.max_samples,
+            bootstrap=self.bootstrap,
+            verbose=True,
+            n_jobs=n_jobs,
+        )
         self.cv_score = np.mean(
-            -1.0 * cross_val_score(
-                pipeline, X_seed, y_seed, cv=KFold(3, shuffle=True),
-                scoring='neg_mean_absolute_error')
+            -1.0
+            * cross_val_score(
+                pipeline,
+                X_seed,
+                y_seed,
+                cv=KFold(3, shuffle=True),
+                scoring="neg_mean_absolute_error",
+            )
         )
         bag_reg.fit(X_seed, y_seed)
 
@@ -573,11 +748,10 @@ class BaggedGaussianProcessStabilityAgent(StabilityAgent):
         expected = preds - stds * self.alpha
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Find the most stable ones up to n_query within hull_distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
         return within_hull.head(self.n_query)
 
@@ -610,11 +784,22 @@ class AgentStabilityAdaBoost(StabilityAgent):
     exploitation in each iteration.  It will exploit a fraction
     of N_query options (frac), and explore the rest of its budget.
     """
-    def __init__(self, candidate_data=None, seed_data=None, n_query=1,
-                 hull_distance=0.0, parallel=cpu_count(),
-                 model=None, uncertainty=True, alpha=0.5,
-                 n_estimators=10, exploit_fraction=0.5,
-                 diversify=False, dynamic_alpha=False):
+
+    def __init__(
+        self,
+        candidate_data=None,
+        seed_data=None,
+        n_query=1,
+        hull_distance=0.0,
+        parallel=cpu_count(),
+        model=None,
+        uncertainty=True,
+        alpha=0.5,
+        n_estimators=10,
+        exploit_fraction=0.5,
+        diversify=False,
+        dynamic_alpha=False,
+    ):
         """
         Args:
             candidate_data (DataFrame): data about the candidates
@@ -642,9 +827,11 @@ class AgentStabilityAdaBoost(StabilityAgent):
         """
 
         super(AgentStabilityAdaBoost, self).__init__(
-            candidate_data=candidate_data, seed_data=seed_data,
-            n_query=n_query, hull_distance=hull_distance,
-            parallel=parallel
+            candidate_data=candidate_data,
+            seed_data=seed_data,
+            n_query=n_query,
+            hull_distance=hull_distance,
+            parallel=parallel,
         )
         self.model = model
         self.exploit_fraction = exploit_fraction
@@ -655,18 +842,35 @@ class AgentStabilityAdaBoost(StabilityAgent):
         self.dynamic_alpha = dynamic_alpha
 
     def get_hypotheses(self, candidate_data, seed_data=None):
+        """
+        Get hypotheses method for AgentStabilityAdaBoost
+
+        Args:
+            candidate_data (pandas.DataFrame): dataframe of candidates
+            seed_data (pandas.DataFrame): dataframe of prior data on
+                which to fit model
+
+        Returns:
+            (pandas.DataFrame): top candidates from the algorithm
+
+        """
         X_cand, X_seed, y_seed = self.update_data(candidate_data, seed_data)
 
-        steps = [('scaler', StandardScaler()), ('ML', self.model)]
+        steps = [("scaler", StandardScaler()), ("ML", self.model)]
         pipeline = Pipeline(steps)
 
         adaboost = AdaBoostRegressor(
-            base_estimator=pipeline, n_estimators=self.n_estimators)
+            base_estimator=pipeline, n_estimators=self.n_estimators
+        )
 
         cv_score = cross_val_score(
-            adaboost, X_seed, y_seed,
-            cv=KFold(3, shuffle=True), scoring='neg_mean_absolute_error')
-        self.cv_score = np.mean(cv_score)*-1
+            adaboost,
+            X_seed,
+            y_seed,
+            cv=KFold(3, shuffle=True),
+            scoring="neg_mean_absolute_error",
+        )
+        self.cv_score = np.mean(cv_score) * -1
 
         # We will take standard scaler out of the pipleine for
         # prediction purposes (we want a single scaler)
@@ -674,35 +878,37 @@ class AgentStabilityAdaBoost(StabilityAgent):
         scaler = StandardScaler()
         X_scaled = scaler.fit_transform(X_seed)
         overall_adaboost = AdaBoostRegressor(
-            base_estimator=self.model,
-            n_estimators=self.n_estimators
+            base_estimator=self.model, n_estimators=self.n_estimators
         )
-        overall_adaboost.fit(X_scaled, self.seed_data['delta_e'])
+        overall_adaboost.fit(X_scaled, self.seed_data["delta_e"])
 
         X_cand = scaler.transform(X_cand)
         expected = overall_adaboost.predict(X_cand)
 
         if self.uncertainty:
-            if self.dynamic_alpha and os.path.exists('iteration.json'):
-                with open('iteration.json', 'r') as f:
+            if self.dynamic_alpha and os.path.exists("iteration.json"):
+                with open("iteration.json", "r") as f:
                     iter = json.load(f)
                 print("dynamic alpha activated")
-                expected -= min(0.1*iter,self.alpha) * self._get_unc_ada(overall_adaboost, X_cand)
+                expected -= min(0.1 * iter, self.alpha) * self._get_unc_ada(
+                    overall_adaboost, X_cand
+                )
             else:
                 expected -= self.alpha * self._get_unc_ada(overall_adaboost, X_cand)
 
         # Update candidate data dataframe with predictions
-        self.update_candidate_stabilities(
-            expected, sort=True, floor=-6.0)
+        self.update_candidate_stabilities(expected, sort=True, floor=-6.0)
 
         # Filter by stability according to hull distance
-        stability_filter = self.candidate_data['pred_stability'] <= self.hull_distance
+        stability_filter = self.candidate_data["pred_stability"] <= self.hull_distance
         within_hull = self.candidate_data[stability_filter]
 
         # Exploitation part:
         n_exploitation = int(self.n_query * self.exploit_fraction)
         if self.diversify:
-            to_compute = diverse_quant( within_hull.index.tolist(), n_exploitation, self.candidate_data)
+            to_compute = diverse_quant(
+                within_hull.index.tolist(), n_exploitation, self.candidate_data
+            )
         else:
             to_compute = within_hull.head(n_exploitation).index.tolist()
         remaining = within_hull.tail(len(within_hull) - n_exploitation)
@@ -727,7 +933,9 @@ class AgentStabilityAdaBoost(StabilityAgent):
         stds = []
         for i in preds:
             average = np.average(i, weights=ada.estimator_weights_)
-            _std = np.sqrt(np.average((i - average) ** 2, weights=ada.estimator_weights_))
+            _std = np.sqrt(
+                np.average((i - average) ** 2, weights=ada.estimator_weights_)
+            )
             stds.append(_std)
         return np.array(stds)
 
@@ -782,14 +990,20 @@ def diverse_quant(points, target_length, df, quantiles=None):
         _df = df.sample(6000)
     else:
         _df = df
-    drop_columns = ['Composition', 'N_species', 'delta_e',
-                    'pred_delta_e', 'pred_stability']
+    drop_columns = [
+        "Composition",
+        "N_species",
+        "delta_e",
+        "pred_delta_e",
+        "pred_stability",
+        "structure"
+    ]
     scaler = StandardScaler()
-    X = scaler.fit_transform(_df.drop(drop_columns, axis=1, errors='ignore'))
+    X = scaler.fit_transform(_df.drop(drop_columns, axis=1, errors="ignore"))
     flatD = pairwise_distances(X).flatten()
 
     _df2 = df.loc[points]
-    X = scaler.transform(_df2.drop(drop_columns, axis=1, errors='ignore'))
+    X = scaler.transform(_df2.drop(drop_columns, axis=1, errors="ignore"))
     D = pairwise_distances(X)
 
     remove_len = len(points) - target_length
@@ -798,7 +1012,7 @@ def diverse_quant(points, target_length, df, quantiles=None):
         q = np.quantile(flatD, alpha)
         to_remove = []
         for i in range(0, len(points)):
-            for j in range(i+1, len(points)):
+            for j in range(i + 1, len(points)):
                 if D[i, j] < q:
                     to_remove.append(j)
         _rl = len(set(to_remove))
@@ -814,5 +1028,8 @@ def diverse_quant(points, target_length, df, quantiles=None):
             d[i] = None
         final_remove_list = list(d.keys())
 
-        return [p for p in points if np.where(_df2.index==p)[0][0]
-                not in final_remove_list][:target_length]
+        return [
+            p
+            for p in points
+            if np.where(_df2.index == p)[0][0] not in final_remove_list
+        ][:target_length]
